#
# Copyright 2018 PyWren Team
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
"""
Generic utility functions for serialization
"""

import base64
import os

try:
    import glob2
except:
    from pywren_ibm_cloud.libs import glob2


def bytes_to_b64str(byte_data):
    byte_data_64 = base64.b64encode(byte_data)
    byte_data_64_ascii = byte_data_64.decode('ascii')
    return byte_data_64_ascii


def create_mod_data(mod_paths):

    module_data = {}
    # load mod paths
    for m in mod_paths:
        if os.path.isdir(m):
            files = glob2.glob(os.path.join(m, "**/*.py"))
            pkg_root = os.path.abspath(os.path.dirname(m))
        else:
            pkg_root = os.path.abspath(os.path.dirname(m))
            files = [m]
        for f in files:
            f = os.path.abspath(f)
            with open(f, 'rb') as file:
                mod_str = file.read()

            dest_filename = f[len(pkg_root)+1:]
            module_data[dest_filename] = bytes_to_b64str(mod_str)

    return module_data



#
# Copyright 2018 PyWren Team
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import logging
from io import BytesIO as StringIO
from pywren_ibm_cloud.libs.cloudpickle import CloudPickler
from pywren_ibm_cloud.serialize.module_dependency import ModuleDependencyAnalyzer

logger = logging.getLogger(__name__)


class SerializeIndependent:

    def __init__(self, preinstalls):
        self.preinstalled_modules = preinstalls
        self.preinstalled_modules.append(['pywren_ibm_cloud', True])
        self._modulemgr = None

    def __call__(self, list_of_objs, **kwargs):
        """
        Serialize f, args, kwargs independently
        """
        self._modulemgr = ModuleDependencyAnalyzer()
        preinstalled_modules = [name for name, _ in self.preinstalled_modules]
        self._modulemgr.ignore(preinstalled_modules)

        cps = []
        strs = []
        for obj in list_of_objs:
            file = StringIO()
            try:
                cp = CloudPickler(file)
                cp.dump(obj)
                cps.append(cp)
                strs.append(file.getvalue())
            finally:
                file.close()

        if '_ignore_module_dependencies' in kwargs:
            ignore_modulemgr = kwargs['_ignore_module_dependencies']
            del kwargs['_ignore_module_dependencies']
        else:
            ignore_modulemgr = False

        if not ignore_modulemgr:
            # Add modules
            for cp in cps:
                for module in cp.modules:
                    self._modulemgr.add(module.__name__)

        mod_paths = self._modulemgr.get_and_clear_paths()
        logger.debug("Modules to transmit: {}".format(None if not mod_paths else mod_paths))

        return (strs, mod_paths)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import enum
import pickle
import logging
from six import reraise
from pywren_ibm_cloud.storage import storage, storage_utils
from pywren_ibm_cloud.libs.tblib import pickling_support

pickling_support.install()

logger = logging.getLogger(__name__)


class JobState(enum.Enum):
    new = 1
    invoked = 2
    running = 3
    ready = 4
    success = 5
    futures = 6
    error = 7


class ResponseFuture:

    """
    Object representing the result of a PyWren invocation. Returns the status of the
    execution and the result when available.
    """
    GET_RESULT_SLEEP_SECS = 2
    GET_RESULT_MAX_RETRIES = 5

    def __init__(self, call_id, callgroup_id, executor_id, activation_id, invoke_metadata, storage_config):
        self.call_id = call_id
        self.callgroup_id = callgroup_id
        self.executor_id = executor_id
        self.activation_id = activation_id

        self.storage_config = storage_config

        self._state = JobState.new
        self._exception = Exception()
        self._return_val = None
        self._new_futures = None
        self._traceback = None
        self._call_invoker_result = None
        self._invoke_metadata = invoke_metadata.copy()

        self.run_status = None
        self.invoke_status = None

        self.status_query_count = 0
        self.output_query_count = 0

        self.storage_path = storage_utils.get_storage_path(self.storage_config)

    def _set_state(self, new_state):
        # FIXME add state machine
        self._state = new_state

    def cancel(self):
        raise NotImplementedError("Cannot cancel dispatched jobs")

    def cancelled(self):
        raise NotImplementedError("Cannot cancel dispatched jobs")

    def running(self):
        raise NotImplementedError()

    @property
    def futures(self):
        """
        The response of a call was a FutureResponse instance.
        It has to wait to the new invocation output.
        """
        return self._state == JobState.futures

    @property
    def done(self):
        if self._state in [JobState.success, JobState.futures, JobState.error]:
            return True
        return False

    @property
    def ready(self):
        if self._state in [JobState.ready, JobState.futures, JobState.error]:
            return True
        return False

    def status(self, check_only=False, throw_except=True, internal_storage=None):
        """
        Return the status returned by the call.
        If the call raised an exception, this method will raise the same exception
        If the future is cancelled before completing then CancelledError will be raised.

        :param check_only: Return None immediately if job is not complete. Default False.
        :param throw_except: Reraise exception if call raised. Default true.
        :param storage_handler: Storage handler to poll cloud storage. Default None.
        :return: Result of the call.
        :raises CancelledError: If the job is cancelled before completed.
        :raises TimeoutError: If job is not complete after `timeout` seconds.
        """
        if self._state == JobState.ready:
            return self.run_status

        if internal_storage is None:
            internal_storage = storage.InternalStorage(self.storage_config)

        storage_utils.check_storage_path(internal_storage.get_storage_config(), self.storage_path)
        call_status = internal_storage.get_call_status(self.executor_id, self.callgroup_id, self.call_id)
        self.status_query_count += 1

        if check_only is True:
            if call_status is None:
                return None

        while call_status is None:
            time.sleep(self.GET_RESULT_SLEEP_SECS)
            call_status = internal_storage.get_call_status(self.executor_id, self.callgroup_id, self.call_id)
            self.status_query_count += 1

        self._invoke_metadata['status_done_timestamp'] = time.time()
        self._invoke_metadata['status_query_count'] = self.status_query_count

        self.run_status = call_status  # this is the remote status information
        self.invoke_status = self._invoke_metadata  # local status information

        total_time = format(round(call_status['end_time'] - call_status['start_time'], 2), '.2f')

        if call_status['exception'] is not None:
            # the wrenhandler had an exception
            self._set_state(JobState.error)
            exception_str = call_status['exception']
            exception_args = call_status['exception_args']

            log_msg = ('Executor ID {} Error in {} {} - Time: {} '
                       'seconds- Result: {}'.format(self.executor_id,
                                                    self.call_id,
                                                    self.activation_id,
                                                    str(total_time),
                                                    exception_args[0]+" "+exception_args[1]))
            logger.debug(log_msg)

            if exception_args[0] == "WRONGVERSION":
                if throw_except:
                    raise Exception("Pywren version mismatch: remote "
                                    "expected version {}, local library is version {}".format(
                                     exception_args[2], exception_args[3]))
                return None
            elif exception_args[0] == "OUTATIME":
                if throw_except:
                    raise Exception("Process ran out of time - {} - {}".format(self.call_id,
                                                                               self.activation_id))
                return None
            elif exception_args[0] == "OUTOFMEMORY":
                if throw_except:
                    raise Exception("Process exceeded maximum memory and was "
                                    "killed - {} - {}".format(self.call_id, self.activation_id))
                return None
            else:
                if 'exception_traceback' in call_status:
                    self._exception = Exception(exception_str, *exception_args)
                if throw_except:
                    raise self._exception
                return None

        log_msg = ('Executor ID {} Response from Function {} - Activation '
                   'ID: {} - Time: {} seconds'.format(self.executor_id,
                                                      self.call_id,
                                                      self.activation_id,
                                                      str(total_time)))
        logger.debug(log_msg)
        self._set_state(JobState.ready)

        if 'new_futures' in call_status:
            _, total_new_futures = call_status['new_futures'].split('/')
            if int(total_new_futures) > 0:
                self.result(check_only, throw_except, internal_storage)

    def result(self, check_only=False, throw_except=True, internal_storage=None):
        """
        Return the value returned by the call.
        If the call raised an exception, this method will raise the same exception
        If the future is cancelled before completing then CancelledError will be raised.

        :param throw_except: Reraise exception if call raised. Default true.
        :param storage_handler: Storage handler to poll cloud storage. Default None.
        :return: Result of the call.
        :raises CancelledError: If the job is cancelled before completed.
        :raises TimeoutError: If job is not complete after `timeout` seconds.
        """
        if self._state == JobState.new:
            raise ValueError("job not yet invoked")

        if self._state == JobState.success:
            return self._return_val

        if self._state == JobState.futures:
            return self._new_futures

        if self._state == JobState.error:
            if throw_except:
                raise self._exception
            else:
                return None

        if internal_storage is None:
            internal_storage = storage.InternalStorage(self.storage_config)

        self.status(check_only, throw_except, internal_storage)

        call_output_time = time.time()
        call_invoker_result = internal_storage.get_call_output(self.executor_id, self.callgroup_id, self.call_id)
        self.output_query_count += 1

        while call_invoker_result is None and self.output_query_count < self.GET_RESULT_MAX_RETRIES:
            time.sleep(self.GET_RESULT_SLEEP_SECS)
            call_invoker_result = internal_storage.get_call_output(self.executor_id, self.callgroup_id, self.call_id)
            self.output_query_count += 1

        if call_invoker_result is None:
            if throw_except:
                raise Exception('Unable to get the output of the function - Activation ID: {}'.format(self.activation_id))
            else:
                self._set_state(JobState.error)
                return None

        call_invoker_result = pickle.loads(call_invoker_result)
        call_output_time_done = time.time()
        self._call_invoker_result = call_invoker_result

        self._invoke_metadata['download_output_time'] = call_output_time_done - call_output_time
        self._invoke_metadata['output_query_count'] = self.output_query_count
        self._invoke_metadata['download_output_timestamp'] = call_output_time_done
        call_success = call_invoker_result['success']
        self.invoke_status = self._invoke_metadata  # local status information

        if call_success:
            log_msg = ('Executor ID {} Got output from Function {} - Activation '
                       'ID: {}'.format(self.executor_id, self.call_id, self.activation_id))
            logger.debug(log_msg)

            function_result = call_invoker_result['result']

            if isinstance(function_result, ResponseFuture):
                self._new_futures = [function_result]
                self._set_state(JobState.futures)
                self.invoke_status['status_done_timestamp'] = self.invoke_status['download_output_timestamp']
                del self.invoke_status['download_output_timestamp']
                return self._new_futures

            elif type(function_result) == list and len(function_result) > 0 \
               and isinstance(function_result[0], ResponseFuture):
                self._new_futures = function_result
                self._set_state(JobState.futures)
                self.invoke_status['status_done_timestamp'] = self.invoke_status['download_output_timestamp']
                del self.invoke_status['download_output_timestamp']
                return self._new_futures

            else:
                self._return_val = function_result
                self._set_state(JobState.success)
                return self._return_val

        elif throw_except:
            self._exception = call_invoker_result['result']
            self._traceback = (call_invoker_result['exc_type'],
                               call_invoker_result['exc_value'],
                               call_invoker_result['exc_traceback'])

            self._set_state(JobState.error)
            if call_invoker_result.get('pickle_fail', False):
                fault = Exception(call_invoker_result['exc_value'])
                reraise(Exception, fault, call_invoker_result['exc_traceback'])
            else:
                reraise(*self._traceback)
        else:
            self._set_state(JobState.error)
            return None  # nothing, don't raise, no value



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import logging
from pywren_ibm_cloud import wrenconfig
from pywren_ibm_cloud.utils import version_str
from pywren_ibm_cloud.storage import storage
from pywren_ibm_cloud.cf_connector import CloudFunctions

logger = logging.getLogger(__name__)

ZIP_LOCATION = os.getcwd()+'/ibmcf_pywren.zip'


def _get_pywren_location():
    my_location = os.path.dirname(os.path.abspath(__file__))
    pw_location = os.path.join(my_location, '..')
    return pw_location


def _create_zip_action():
    pywren_location = _get_pywren_location()

    cmd = 'cd ' + pywren_location + '/..; zip -FSr "' + ZIP_LOCATION + '" pywren_ibm_cloud/ -x "*__pycache__*"'
    res = os.popen(cmd)
    result = res.read()
    logger.debug(result)
    code = res.close()
    if code:
        raise Exception('Unable to create the {} action package'.format(ZIP_LOCATION))

    cmd = 'cd ' + pywren_location + '/action; zip -r "' + ZIP_LOCATION + '" __main__.py'
    res = os.popen(cmd)
    result = res.read()
    logger.debug(result)
    code = res.close()
    if code:
        raise Exception('Unable to add __main__.py into the {} action package'.format(ZIP_LOCATION))


def _extract_modules(image_name, memory, cf_client, config):
    # Extract installed Python modules from docker image
    # And store them into storage

    # Create storage_handler to upload modules file
    storage_config = wrenconfig.extract_storage_config(config)
    internal_storage = storage.InternalStorage(storage_config)

    pywren_location = _get_pywren_location()
    action_location = os.path.join(pywren_location, "runtime", "extract_modules.py")

    with open(action_location, "r") as action_py:
        action_code = action_py.read()

    modules_action_name = '{}_modules'.format(image_name.replace('/', '@').replace(':', '_'))

    old_stdout = sys.stdout
    sys.stdout = open(os.devnull, 'w')
    cf_client.create_action(modules_action_name, image_name, code=action_code, is_binary=False)
    sys.stdout = old_stdout

    memory = cf_client.default_runtime_memory if not memory else memory
    runtime_name = '{}_{}'.format(image_name, memory)
    runtime_meta = cf_client.invoke_with_result(modules_action_name)
    internal_storage.put_runtime_info(runtime_name, runtime_meta)
    cf_client.delete_action(modules_action_name)


def _create_blackbox_runtime(image_name, memory, cf_client):
    # Create runtime_name from image_name
    memory = cf_client.default_runtime_memory if not memory else memory
    action_name = image_name.replace('/', '@').replace(':', '_')
    action_name = '{}_{}'.format(action_name, memory)

    # Upload zipped PyWren action
    with open(ZIP_LOCATION, "rb") as action_zip:
        action_bin = action_zip.read()
    cf_client.create_action(action_name, image_name, code=action_bin, memory=memory)


def create_runtime(image_name, memory=None, config=None):
    print('Creating a new docker image from the Dockerfile')
    print('Docker image name: {}'.format(image_name))

    cmd = 'docker build -t {} .'.format(image_name)
    res = os.system(cmd)
    if res != 0:
        exit()

    cmd = 'docker push {}'.format(image_name)
    res = os.system(cmd)
    if res != 0:
        exit()

    if config is None:
        config = wrenconfig.default()
    else:
        config = wrenconfig.default(config)

    cf_config = wrenconfig.extract_cf_config(config)
    cf_client = CloudFunctions(cf_config)
    cf_client.create_package()
    _create_zip_action()
    _extract_modules(image_name, memory, cf_client, config)
    _create_blackbox_runtime(image_name, memory, cf_client)


def clone_runtime(image_name, memory=None, config=None):
    print('Cloning docker image {}'.format(image_name))

    if config is None:
        config = wrenconfig.default()
    else:
        config = wrenconfig.default(config)

    cf_config = wrenconfig.extract_cf_config(config)
    cf_client = CloudFunctions(cf_config)
    cf_client.create_package()
    _create_zip_action()
    _extract_modules(image_name,  memory, cf_client, config)
    _create_blackbox_runtime(image_name, memory, cf_client)


def update_runtime(image_name, memory=None, config=None):
    print('Updating runtime: {}'.format(image_name))
    if config is None:
        config = wrenconfig.default()
    else:
        config = wrenconfig.default(config)

    cf_config = wrenconfig.extract_cf_config(config)
    cf_client = CloudFunctions(cf_config)

    _create_zip_action()
    _create_blackbox_runtime(image_name, memory, cf_client)


def delete_runtime(image_name, memory=None, config=None):
    print('Deleting runtime: {}'.format(image_name))

    if config is None:
        config = wrenconfig.default()
    else:
        config = wrenconfig.default(config)

    storage_config = wrenconfig.extract_storage_config(config)
    storage_client = storage.InternalStorage(storage_config)
    cf_config = wrenconfig.extract_cf_config(config)
    cf_client = CloudFunctions(cf_config)

    memory = cf_client.default_runtime_memory if not memory else memory

    runtime_name = '{}_{}'.format(image_name, memory)
    storage_client.delete_runtime_info(runtime_name)

    action_name = runtime_name.replace('/', '@').replace(':', '_')
    cf_client.delete_action(action_name)


def deploy_default_runtime(memory=None, config=None):
    this_version_str = version_str(sys.version_info)
    print('Updating default Python {} runtime'.format(this_version_str))
    if config is None:
        config = wrenconfig.default()
    else:
        config = wrenconfig.default(config)

    if this_version_str == '3.5':
        image_name = wrenconfig.RUNTIME_DEFAULT_35
    elif this_version_str == '3.6':
        image_name = wrenconfig.RUNTIME_DEFAULT_36
    elif this_version_str == '3.7':
        image_name = wrenconfig.RUNTIME_DEFAULT_37

    cf_config = wrenconfig.extract_cf_config(config)
    cf_client = CloudFunctions(cf_config)
    cf_client.create_package()

    _create_zip_action()
    _extract_modules(image_name, memory, cf_client, config)
    _create_blackbox_runtime(image_name, memory, cf_client)



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import requests
import base64
import os
import json
import ssl
from urllib.parse import urlparse
import http.client
import logging
import time
from pywren_ibm_cloud.version import __version__

logger = logging.getLogger(__name__)


class CloudFunctions:

    def __init__(self, config):
        """
        Constructor
        """
        self.api_key = str.encode(config['api_key'])
        self.endpoint = config['endpoint'].replace('http:', 'https:')
        self.namespace = config['namespace']
        self.default_runtime_memory = int(config['runtime_memory'])
        self.default_runtime_timeout = int(config['runtime_timeout'])
        self.is_cf_cluster = config['is_cf_cluster']
        self.package = 'pywren_v'+__version__

        auth = base64.encodebytes(self.api_key).replace(b'\n', b'')
        self.session = requests.session()
        default_user_agent = self.session.headers['User-Agent']
        self.headers = {
            'content-type': 'application/json',
            'Authorization': 'Basic %s' % auth.decode('UTF-8'),
            'User-Agent': default_user_agent + ' pywren-ibm-cloud'
        }

        self.session.headers.update(self.headers)
        adapter = requests.adapters.HTTPAdapter()
        self.session.mount('https://', adapter)

        msg = 'IBM Cloud Functions init for'
        logger.info('{} namespace: {}'.format(msg, self.namespace))
        logger.info('{} host: {}'.format(msg, self.endpoint))
        logger.debug("CF user agent set to: {}".format(self.session.headers['User-Agent']))

        if logger.getEffectiveLevel() == logging.WARNING:
            print("{} Namespace: {}".format(msg, self.namespace))
            print("{} Host: {}".format(msg, self.endpoint))

    def create_action(self, action_name, image_name, code=None, memory=None, kind='blackbox',
                      is_binary=True, overwrite=True):
        """
        Create an IBM Cloud Function
        """
        data = {}
        limits = {}
        cfexec = {}
        limits['memory'] = self.default_runtime_memory if not memory else memory
        limits['timeout'] = self.default_runtime_timeout
        data['limits'] = limits

        cfexec['kind'] = kind
        if kind == 'blackbox':
            cfexec['image'] = image_name
        cfexec['binary'] = is_binary
        cfexec['code'] = base64.b64encode(code).decode("utf-8") if is_binary else code
        data['exec'] = cfexec

        logger.debug('I am about to create a new cloud function action: {}'.format(action_name))
        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'actions', self.package,
                           action_name + "?overwrite=" + str(overwrite)).replace("\\", "/")
        res = self.session.put(url, json=data)
        resp_text = res.json()

        if res.status_code == 200:
            print("OK --> Created action {}".format(action_name))
        else:
            msg = 'An error occurred creating/updating action {}: {}'.format(action_name, resp_text['error'])
            raise Exception(msg)

    def get_action(self, action_name):
        """
        Get an IBM Cloud Function
        """
        logger.debug("I am about to get a cloud function action: {}".format(action_name))
        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'actions', self.package, action_name).replace("\\", "/")
        res = self.session.get(url)
        return res.json()

    def delete_action(self, action_name):
        """
        Delete an IBM Cloud Function
        """
        logger.debug("Delete cloud function action: {}".format(action_name))

        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'actions', self.package, action_name).replace("\\", "/")
        res = self.session.delete(url)
        resp_text = res.json()

        if res.status_code != 200:
            logger.debug('An error occurred deleting action {}: {}'.format(action_name, resp_text['error']))

    def update_memory(self, action_name, memory):
        logger.debug('I am about to update the memory of the {} action to {}'.format(action_name, memory))
        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'actions', self.package,
                           action_name + "?overwrite=True").replace("\\", "/")

        data = {"limits": {"memory": memory}}
        res = self.session.put(url, json=data)
        resp_text = res.json()

        if res.status_code != 200:
            logger.debug('An error occurred updating action {}: {}'.format(action_name, resp_text['error']))
        else:
            logger.debug("OK --> Updated action memory {}".format(action_name))

    def create_package(self):
        logger.debug('I am about to crate the package {}'.format(self.package))
        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'packages',
                           self.package + "?overwrite=False").replace("\\", "/")

        data = {"name": self.package}
        res = self.session.put(url, json=data)
        resp_text = res.json()

        if res.status_code != 200:
            logger.debug('Package {}: {}'.format(self.package, resp_text['error']))
        else:
            logger.debug("OK --> Created package {}".format(self.package))

    def invoke(self, action_name, payload):
        """
        Wrapper around actual invocation methods
        """
        if self.is_cf_cluster:
            return self.internal_invoke(action_name, payload)
        else:
            return self.remote_invoke(action_name, payload)

    def remote_invoke(self, action_name, payload):
        """
        Invoke an IBM Cloud Function. Better from a remote network.
        """
        exec_id = payload['executor_id']
        call_id = payload['call_id']
        url = os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                           self.namespace, 'actions', self.package, action_name).replace("\\", "/")

        try:
            resp = self.session.post(url, json=payload)
            data = resp.json()
            resp_time = format(round(resp.elapsed.total_seconds(), 3), '.3f')
        except Exception:
            return self.remote_invoke(action_name, payload)

        if 'activationId' in data:
            log_msg = ('Executor ID {} Function {} - Activation ID: '
                       '{} - Time: {} seconds'.format(exec_id, call_id,
                                                      data["activationId"],
                                                      resp_time))
            logger.debug(log_msg)
            return data["activationId"]
        else:
            logger.debug(data)
            return None

    def internal_invoke(self, action_name, payload):
        """
        Invoke an IBM Cloud Function. Better from a cloud function.
        """
        exec_id = payload['executor_id']
        call_id = payload['call_id']

        url = urlparse(os.path.join(self.endpoint, 'api', 'v1', 'namespaces',
                                    self.namespace, 'actions', self.package, action_name))
        ctx = ssl._create_unverified_context()

        try:
            start = time.time()
            conn = http.client.HTTPSConnection(url.netloc, context=ctx)
            conn.request("POST", url.geturl(),
                         body=json.dumps(payload),
                         headers=self.headers)
            resp = conn.getresponse()
            data = resp.read()
            roundtrip = time.time() - start
            resp_time = format(round(roundtrip, 3), '.3f')
            data = json.loads(data.decode("utf-8"))
            conn.close()
        except Exception:
            conn.close()
            return self.internal_invoke(action_name, payload)

        if 'activationId' in data:
            log_msg = ('Executor ID {} Function {} - Activation ID: '
                       '{} - Time: {} seconds'.format(exec_id, call_id,
                                                      data["activationId"],
                                                      resp_time))
            logger.debug(log_msg)
            return data["activationId"]
        else:
            logger.debug(data)
            return None

    def invoke_with_result(self, action_name, payload={}):
        """
        Invoke an IBM Cloud Function waiting for the result.
        """
        url = os.path.join(self.endpoint, 'api', 'v1',
                           'namespaces', self.namespace, 'actions', self.package,
                           action_name + "?blocking=true&result=true").replace("\\", "/")
        resp = self.session.post(url, json=payload)
        result = resp.json()

        return result



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import time
import logging
import random
from pywren_ibm_cloud.cf_connector import CloudFunctions
from pywren_ibm_cloud.wrenconfig import extract_cf_config

logger = logging.getLogger(__name__)


class IBMCloudFunctionsInvoker:

    def __init__(self, config):
        self.log_level = os.getenv('PYWREN_LOG_LEVEL')
        cf_config = extract_cf_config(config)
        self.namespace = cf_config['namespace']
        self.endpoint = cf_config['endpoint']
        self.runtime = cf_config['runtime']
        self.runtime_memory = int(cf_config['runtime_memory'])
        self.runtime_timeout = int(cf_config['runtime_timeout'])

        self.action_name = self.runtime.replace('/', '@').replace(':', '_')
        self.action_name = '{}_{}'.format(self.action_name, self.runtime_memory)

        self.invocation_retry = config['pywren']['invocation_retry']
        self.retry_sleeps = config['pywren']['retry_sleeps']
        self.retries = config['pywren']['retries']

        self.client = CloudFunctions(cf_config)

        log_msg = 'IBM Cloud Functions init for Runtime: {} - {}MB'.format(self.runtime, self.runtime_memory)
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg, end=' ')

    def invoke(self, payload):
        """
        Invoke -- return information about this invocation
        """
        act_id = self.client.invoke(self.action_name, payload)
        attempts = 1

        while not act_id and self.invocation_retry and attempts < self.retries:
            attempts += 1
            selected_sleep = random.choice(self.retry_sleeps)
            exec_id = payload['executor_id']
            call_id = payload['call_id']

            log_msg = ('Executor ID {} Function {} - Invocation failed - retry {} in {} seconds'.format(exec_id, call_id, attempts, selected_sleep))
            logger.debug(log_msg)

            time.sleep(selected_sleep)
            act_id = self.client.invoke(self.action_name, payload)

        return act_id

    def config(self):
        """
        Return config dict
        """
        return {'cf_runtime': self.runtime,
                'cf_runtime_memory': self.runtime_memory,
                'cf_runtime_timeout': self.runtime_timeout,
                'cf_namespace': self.namespace,
                'cf_endpoint': self.endpoint}



try:
    import copy_reg
except ImportError:
    import copyreg as copy_reg
from types import TracebackType

from . import Frame
from . import Traceback


def unpickle_traceback(tb_frame, tb_lineno, tb_next):
    ret = object.__new__(Traceback)
    ret.tb_frame = tb_frame
    ret.tb_lineno = tb_lineno
    ret.tb_next = tb_next
    return ret.as_traceback()


def pickle_traceback(tb):
    return unpickle_traceback, (Frame(tb.tb_frame), tb.tb_lineno, tb.tb_next and Traceback(tb.tb_next))


def install():
    copy_reg.pickle(TracebackType, pickle_traceback)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import json
from pywren_ibm_cloud.version import __version__
from pywren_ibm_cloud.storage.backends.cos import COSBackend
from pywren_ibm_cloud.storage.backends.swift import SwiftBackend
from pywren_ibm_cloud.storage.exceptions import StorageNoSuchKeyError
from pywren_ibm_cloud.storage.storage_utils import create_status_key, create_output_key, status_key_suffix


class InternalStorage:
    """
    An InternalStorage object is used by executors and other components to access underlying storage backend
    without exposing the the implementation details.
    """

    def __init__(self, config):
        self.storage_config = config
        self.backend_type = config['storage_backend']
        self.storage_bucket = config['storage_bucket']
        self.prefix = config['storage_prefix']

        if self.backend_type == 'ibm_cos':
            self.backend_handler = COSBackend(config['ibm_cos'])
        elif self.backend_type == 'swift':
            self.backend_handler = SwiftBackend(config['swift'])
        else:
            raise NotImplementedError(("Using {} as internal storage backend is" +
                                       "not supported yet").format(self.backend_type))

    def get_storage_config(self):
        """
        Retrieves the configuration of this storage handler.
        :return: storage configuration
        """
        return self.storage_config

    def put_data(self, key, data):
        """
        Put data object into storage.
        :param key: data key
        :param data: data content
        :return: None
        """
        return self.backend_handler.put_object(self.storage_bucket, key, data)

    def put_func(self, key, func):
        """
        Put serialized function into storage.
        :param key: function key
        :param func: serialized function
        :return: None
        """
        return self.backend_handler.put_object(self.storage_bucket, key, func)

    def get_data(self, key, stream=False, extra_get_args={}):
        """
        Get data object from storage.
        :param key: data key
        :return: data content
        """
        return self.backend_handler.get_object(self.storage_bucket, key, stream, extra_get_args)

    def get_func(self, key):
        """
        Get serialized function from storage.
        :param key: function key
        :return: serialized function
        """
        return self.backend_handler.get_object(self.storage_bucket, key)

    def get_callset_status(self, executor_id):
        """
        Get the status of a callset.
        :param executor_id: executor's ID
        :return: A list of call IDs that have updated status.
        """
        # TODO: a better API for this is to return status for all calls in the callset. We'll fix
        #  this in scheduler refactoring.
        callset_prefix = os.path.join(self.prefix, executor_id)
        keys = self.backend_handler.list_keys_with_prefix(self.storage_bucket, callset_prefix)
        suffix = status_key_suffix
        status_keys = [k for k in keys if suffix in k]
        call_ids = [tuple(k[len(callset_prefix)+1:].split("/")[:2]) for k in status_keys]
        return call_ids

    def get_call_status(self, executor_id, callgroup_id, call_id):
        """
        Get status of a call.
        :param executor_id: executor ID of the call
        :param call_id: call ID of the call
        :return: A dictionary containing call's status, or None if no updated status
        """
        status_key = create_status_key(self.prefix, executor_id, callgroup_id, call_id)
        try:
            data = self.backend_handler.get_object(self.storage_bucket, status_key)
            return json.loads(data.decode('ascii'))
        except StorageNoSuchKeyError:
            return None

    def get_call_output(self, executor_id, callgroup_id, call_id):
        """
        Get the output of a call.
        :param executor_id: executor ID of the call
        :param call_id: call ID of the call
        :return: Output of the call.
        """
        output_key = create_output_key(self.prefix, executor_id, callgroup_id, call_id)
        try:
            return self.backend_handler.get_object(self.storage_bucket, output_key)
        except StorageNoSuchKeyError:
            return None

    def get_runtime_info(self, runtime_name):
        """
        Get the metadata given a runtime name.
        :param runtime: name of the runtime
        :return: runtime metadata
        """
        key = os.path.join('runtimes', __version__,  runtime_name+".meta.json")
        try:
            json_str = self.backend_handler.get_object(self.storage_bucket, key)
            runtime_meta = json.loads(json_str.decode("ascii"))
            return runtime_meta
        except StorageNoSuchKeyError:
            raise Exception('The runtime {} is not installed.'.format(runtime_name))

    def put_runtime_info(self, runtime_name, runtime_meta):
        """
        Puit the metadata given a runtime config.
        :param runtime: name of the runtime
        :param runtime_meta metadata
        """
        key = os.path.join('runtimes', __version__, runtime_name+".meta.json")
        self.backend_handler.put_object(self.storage_bucket, key, json.dumps(runtime_meta))

    def list_temporal_data(self, executor_id):
        """
        List the temporal data used by PyWren.
        :param bucket: bucket key
        :param prefix: prefix to search for
        :return: list of objects
        """
        return self.backend_handler.list_keys_with_prefix(self.storage_bucket, executor_id)

    def delete_temporal_data(self, key_list):
        """
        Delete temporal data from PyWren.
        :param bucket: bucket name
        :param key: data key
        """
        return self.backend_handler.delete_objects(self.storage_bucket, key_list)



import matplotlib
matplotlib.use('Agg')
import os
import pylab
import logging
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import patches as mpatches
from matplotlib.collections import LineCollection
sns.set_style('whitegrid')
logger = logging.getLogger(__name__)


def create_timeline(dst, name, pw_start_time, run_statuses, invoke_statuses):
    results_df = pd.DataFrame(run_statuses)

    if invoke_statuses:
        invoke_df = pd.DataFrame(invoke_statuses)
        results_df = pd.concat([results_df, invoke_df], axis=1)
        Cols = list(results_df.columns)
        for i, item in enumerate(results_df.columns):
            if item in results_df.columns[:i]:
                Cols[i] = "toDROP"
        results_df.columns = Cols
        results_df = results_df.drop("toDROP", 1)

    palette = sns.color_palette("deep", 6)
    #time_offset = np.min(results_df.host_submit_time)
    fig = pylab.figure(figsize=(10, 6))
    ax = fig.add_subplot(1, 1, 1)
    total_jobs = len(results_df)

    y = np.arange(total_jobs)
    point_size = 10

    fields = [('host submit', results_df.host_submit_time - pw_start_time),
              ('action start', results_df.start_time - pw_start_time),
              #('jobrunner start', results_df.jobrunner_start - pw_start_time),
              ('action done', results_df.end_time - pw_start_time)]

    if 'download_output_timestamp' in results_df:
        fields.append(('results fetched', results_df.download_output_timestamp - pw_start_time))
    elif 'status_done_timestamp' in results_df:
        fields.append(('status fetched', results_df.status_done_timestamp - pw_start_time))

    patches = []
    for f_i, (field_name, val) in enumerate(fields):
        ax.scatter(val, y, c=[palette[f_i]], edgecolor='none', s=point_size, alpha=0.8)
        patches.append(mpatches.Patch(color=palette[f_i], label=field_name))

    ax.set_xlabel('wallclock time (sec)')
    ax.set_ylabel('job')
    #pylab.ylim(0, 10)

    legend = pylab.legend(handles=patches, loc='upper right', frameon=True)
    #pylab.title("Runtime for {} jobs of {:3.0f}M double ops (dgemm) each".format(total_jobs, JOB_GFLOPS))
    legend.get_frame().set_facecolor('#FFFFFF')

    plot_step = int(np.max([1, total_jobs/32]))

    y_ticks = np.arange(total_jobs//plot_step + 2) * plot_step
    ax.set_yticks(y_ticks)
    ax.set_ylim(-0.02*total_jobs, total_jobs*1.05)
    for y in y_ticks:
        ax.axhline(y, c='k', alpha=0.1, linewidth=1)

    if 'download_output_timestamp' in results_df:
        max_seconds = np.max(results_df.download_output_timestamp - pw_start_time)*1.25
    elif 'status_done_timestamp' in results_df:
        max_seconds = np.max(results_df.status_done_timestamp - pw_start_time)*1.25
    else:
        max_seconds = np.max(results_df.end_time - pw_start_time)*1.25

    xplot_step = int(max_seconds/8)
    x_ticks = np.arange(max_seconds//xplot_step + 2) * xplot_step
    ax.set_xlim(0, max_seconds)

    ax.set_xticks(x_ticks)
    for x in x_ticks:
        ax.axvline(x, c='k', alpha=0.2, linewidth=0.8)

    ax.grid(False)
    fig.tight_layout()
    fig.savefig(os.path.join(dst, name+"_timeline.png"))


def create_histogram(dst, name, pw_start_time, run_statuses):
    runtime_bins = np.linspace(0, 600, 600)

    def compute_times_rates(time_rates):
        x = np.array(time_rates)

        #tzero = np.min(x[:, 0])
        tzero = pw_start_time
        start_time = x[:, 0] - tzero
        end_time = x[:, 1] - tzero

        N = len(start_time)

        runtime_jobs_hist = np.zeros((N, len(runtime_bins)))

        for i in range(N):
            s = start_time[i]
            e = end_time[i]
            a, b = np.searchsorted(runtime_bins, [s, e])
            if b-a > 0:
                runtime_jobs_hist[i, a:b] = 1

        return {'start_time': start_time,
                'end_time': end_time,
                'runtime_jobs_hist': runtime_jobs_hist}

    fig = pylab.figure(figsize=(10, 6))
    ax = fig.add_subplot(1, 1, 1)

    time_rates = [(rs['start_time'], rs['end_time']) for rs in run_statuses]

    time_hist = compute_times_rates(time_rates)

    N = len(time_hist['start_time'])
    line_segments = LineCollection([[[time_hist['start_time'][i], i],
                                     [time_hist['end_time'][i], i]] for i in range(N)],
                                   linestyles='solid', color='k', alpha=0.6, linewidth=0.4)

    ax.add_collection(line_segments)

    ax.plot(runtime_bins, time_hist['runtime_jobs_hist'].sum(axis=0),
            label='active jobs total', zorder=-1)

    #ax.set_xlim(0, x_lim)
    ax.set_xlim(0, np.max(time_hist['end_time'])*3)
    ax.set_ylim(0, len(time_hist['start_time'])*1.05)
    ax.set_xlabel("time (sec)")

    ax.set_ylabel("IBM Cloud function execution")
    ax.grid(False)
    ax.legend(loc='upper right')

    fig.tight_layout()
    fig.savefig(os.path.join(dst, name+"_histogram.png"))



from pywren_ibm_cloud.runtime.deploy_utils import create_runtime
from pywren_ibm_cloud.runtime.deploy_utils import clone_runtime
from pywren_ibm_cloud.runtime.deploy_utils import deploy_default_runtime
from pywren_ibm_cloud.runtime.deploy_utils import update_runtime
from pywren_ibm_cloud.runtime.metadata import get_runtime_preinstalls



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from pywren_ibm_cloud import wrenlogging
from pywren_ibm_cloud.action.handler import ibm_cloud_function_handler

wrenlogging.ow_config(logging.INFO)
logger = logging.getLogger('__main__')


def main(args):
    logger.info("Starting IBM Cloud Function execution")
    ibm_cloud_function_handler(args)
    return {"greeting": "Finished"}



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import time
import pickle
import logging
import inspect
import pywren_ibm_cloud as pywren
import pywren_ibm_cloud.version as version
import pywren_ibm_cloud.utils as wrenutil
import pywren_ibm_cloud.wrenconfig as wrenconfig
from pywren_ibm_cloud.wait import wait
from multiprocessing.pool import ThreadPool
from pywren_ibm_cloud.future import ResponseFuture, JobState
from pywren_ibm_cloud.runtime import get_runtime_preinstalls
from pywren_ibm_cloud.storage.backends.cos import COSBackend
from pywren_ibm_cloud.serialize import serialize, create_mod_data
from pywren_ibm_cloud.storage.storage_utils import create_keys, create_func_key, create_agg_data_key
from pywren_ibm_cloud.partitioner import create_partitions, partition_processor


logger = logging.getLogger(__name__)


class Executor(object):

    def __init__(self, invoker, config, internal_storage):
        self.log_level = os.getenv('PYWREN_LOG_LEVEL')
        self.invoker = invoker
        self.config = config
        self.internal_storage = internal_storage

        self.runtime_name = self.config['pywren']['runtime']
        self.runtime_memory = self.config['pywren']['runtime_memory']
        runtime_preinstalls = get_runtime_preinstalls(self.internal_storage,
                                                      self.runtime_name,
                                                      self.runtime_memory,
                                                      self.config)
        self.serializer = serialize.SerializeIndependent(runtime_preinstalls)

        self.map_item_limit = None
        if 'scheduler' in self.config:
            if 'map_item_limit' in config['scheduler']:
                self.map_item_limit = config['scheduler']['map_item_limit']

        if 'PYWREN_EXECUTOR_ID' in os.environ:
            self.executor_id = os.environ['PYWREN_EXECUTOR_ID']
        else:
            self.executor_id = wrenutil.create_executor_id()

        log_msg = 'IBM Cloud Functions executor created with ID {}'.format(self.executor_id)
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg)

    def invoke_with_keys(self, func_key, data_key, output_key,
                         status_key, executor_id, callgroup_id,
                         call_id, extra_env,
                         extra_meta, data_byte_range,
                         host_job_meta, job_max_runtime,
                         overwrite_invoke_args=None):

        log_level = 'INFO' if not self.log_level else self.log_level

        arg_dict = {
            'config': self.config,
            'log_level': log_level,
            'func_key': func_key,
            'data_key': data_key,
            'output_key': output_key,
            'status_key': status_key,
            'job_max_runtime': job_max_runtime,
            'data_byte_range': data_byte_range,
            'executor_id': executor_id,
            'callgroup_id': callgroup_id,
            'call_id': call_id,
            'pywren_version': version.__version__}

        if extra_env is not None:
            logger.debug("Extra environment vars {}".format(extra_env))
            arg_dict['extra_env'] = extra_env

        if extra_meta is not None:
            # sanity
            for k, v in extra_meta.items():
                if k in arg_dict:
                    raise ValueError("Key {} already in dict".format(k))
                arg_dict[k] = v

        host_submit_time = time.time()
        arg_dict['host_submit_time'] = host_submit_time

        # logger.debug("Executor ID {} Activation {} invoke".format(executor_id, call_id))

        # overwrite explicit args, mostly used for testing via injection
        if overwrite_invoke_args is not None:
            arg_dict.update(overwrite_invoke_args)

        cf_invoke_time_start = time.time()
        # do the invocation
        activation_id = self.invoker.invoke(arg_dict)

        if not activation_id:
            raise ValueError("Executor ID {} Activation {} failed, therefore job is failed".format(executor_id, call_id))

        host_job_meta['cf_activation_id'] = activation_id
        host_job_meta['cf_invoke_timestamp'] = cf_invoke_time_start
        host_job_meta['cf_invoke_time'] = time.time() - cf_invoke_time_start

        # logger.debug("Executor ID {} Activation {} complete".format(executor_id, call_id))

        host_job_meta.update(self.invoker.config())
        host_job_meta.update(arg_dict)
        del host_job_meta['config']
        storage_config = self.internal_storage.get_storage_config()
        fut = ResponseFuture(call_id, callgroup_id, executor_id, activation_id, host_job_meta, storage_config)
        fut._set_state(JobState.invoked)

        return fut

    @staticmethod
    def agg_data(data_strs):
        ranges = []
        pos = 0
        for datum in data_strs:
            l = len(datum)
            ranges.append((pos, pos+l-1))
            pos += l
        return b"".join(data_strs), ranges

    def call_async(self, func, data, extra_env=None, extra_meta=None, runtime_timeout=wrenconfig.RUNTIME_TIMEOUT):
        """
        Wrapper to launch one function invocation.
        """

        return self._map(func, [data], extra_env=extra_env, extra_meta=extra_meta, job_max_runtime=runtime_timeout)

    def map(self, map_function, iterdata, obj_chunk_size=None, extra_env=None, extra_meta=None,
            remote_invocation=False, remote_invocation_groups=100, invoke_pool_threads=128,
            data_all_as_one=True, job_max_runtime=wrenconfig.RUNTIME_TIMEOUT,
            overwrite_invoke_args=None, exclude_modules=None):
        """
        Wrapper to launch map() method.  It integrates COS logic to process objects.
        """
        data = wrenutil.iterdata_as_list(iterdata)
        map_func = map_function
        map_iterdata = data
        new_invoke_pool_threads = invoke_pool_threads
        parts_per_object = None

        if wrenutil.is_object_processing(map_function):
            '''
            If it is object processing function, create partitions according chunk_size
            '''
            logger.debug("Calling map on partitions from object storage flow")
            arg_data = wrenutil.verify_args(map_function, data, object_processing=True)
            storage = COSBackend(self.config['ibm_cos'])
            map_iterdata, parts_per_object = create_partitions(arg_data, obj_chunk_size, storage)
            map_func = partition_processor(map_function)

        # Remote invocation functionality
        if len(iterdata) > 1 and remote_invocation:
            runtime_name = self.runtime_name
            runtime_memory = self.runtime_memory
            rabbitmq_monitor = "PYWREN_RABBITMQ_MONITOR" in os.environ

            def remote_invoker(input_data):
                pw = pywren.ibm_cf_executor(runtime=runtime_name,
                                            runtime_memory=runtime_memory,
                                            rabbitmq_monitor=rabbitmq_monitor)
                return pw.map(map_function, input_data,
                              invoke_pool_threads=invoke_pool_threads,
                              extra_env=extra_env,
                              extra_meta=extra_meta)

            map_func = remote_invoker
            map_iterdata = [[iterdata[x:x+remote_invocation_groups]]
                            for x in range(0, len(iterdata), remote_invocation_groups)]
            new_invoke_pool_threads = 1

        map_futures = self._map(map_func, map_iterdata,
                                extra_env=extra_env,
                                extra_meta=extra_meta,
                                invoke_pool_threads=new_invoke_pool_threads,
                                data_all_as_one=data_all_as_one,
                                overwrite_invoke_args=overwrite_invoke_args,
                                exclude_modules=exclude_modules,
                                original_func_name=map_function.__name__,
                                job_max_runtime=job_max_runtime)

        return map_futures, parts_per_object

    def _map(self, func, iterdata, extra_env=None, extra_meta=None, invoke_pool_threads=128,
             data_all_as_one=True, overwrite_invoke_args=None, exclude_modules=None,
             original_func_name=None, job_max_runtime=wrenconfig.RUNTIME_TIMEOUT):
        """
        :param func: the function to map over the data
        :param iterdata: An iterable of input data
        :param extra_env: Additional environment variables for CF environment. Default None.
        :param extra_meta: Additional metadata to pass to CF. Default None.
        :param remote_invocation: Enable remote invocation. Default False.
        :param invoke_pool_threads: Number of threads to use to invoke.
        :param data_all_as_one: upload the data as a single object. Default True
        :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
        :param exclude_modules: Explicitly keep these modules from pickled dependencies.
        :param original_func_name: Name of the function to invoke.
        :return: A list with size `len(iterdata)` of futures for each job
        :rtype:  list of futures.
        """
        if original_func_name:
            func_name = original_func_name
        else:
            func_name = func.__name__

        data = wrenutil.iterdata_as_list(iterdata)

        if extra_env is not None:
            extra_env = wrenutil.convert_bools_to_string(extra_env)

        if not data:
            return []

        if self.map_item_limit is not None and len(data) > self.map_item_limit:
            raise ValueError("len(data) ={}, exceeding map item limit of {}"
                             "consider mapping over a smaller"
                             "number of items".format(len(data),
                                                      self.map_item_limit))

        # This allows multiple parameters in functions
        data = wrenutil.verify_args(func, data)

        callgroup_id = wrenutil.create_callgroup_id()

        host_job_meta = {}

        pool = ThreadPool(invoke_pool_threads)

        log_msg = 'Executor ID {} Serializing function and data'.format(self.executor_id)
        logger.debug(log_msg)
        # pickle func and all data (to capture module dependencies)
        func_and_data_ser, mod_paths = self.serializer([func] + data)

        func_str = func_and_data_ser[0]
        data_strs = func_and_data_ser[1:]
        data_size_bytes = sum(len(x) for x in data_strs)

        agg_data_key = None
        host_job_meta['agg_data'] = False
        host_job_meta['data_size_bytes'] = data_size_bytes

        log_msg = 'Executor ID {} Uploading function and data'.format(self.executor_id)
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg, end=' ')

        if data_size_bytes < wrenconfig.MAX_AGG_DATA_SIZE and data_all_as_one:
            agg_data_key = create_agg_data_key(self.internal_storage.prefix, self.executor_id, callgroup_id)
            agg_data_bytes, agg_data_ranges = self.agg_data(data_strs)
            agg_upload_time = time.time()
            self.internal_storage.put_data(agg_data_key, agg_data_bytes)
            host_job_meta['agg_data'] = True
            host_job_meta['data_upload_time'] = time.time() - agg_upload_time
            host_job_meta['data_upload_timestamp'] = time.time()
        else:
            log_msg = ('Executor ID {} Total data exceeded '
                       'maximum size of {} bytes'.format(self.executor_id,
                                                         wrenconfig.MAX_AGG_DATA_SIZE))
            logger.warning(log_msg)

        if exclude_modules:
            for module in exclude_modules:
                for mod_path in list(mod_paths):
                    if module in mod_path and mod_path in mod_paths:
                        mod_paths.remove(mod_path)

        module_data = create_mod_data(mod_paths)
        # Create func and upload
        func_module_str = pickle.dumps({'func': func_str, 'module_data': module_data}, -1)
        host_job_meta['func_module_bytes'] = len(func_module_str)

        func_upload_time = time.time()
        func_key = create_func_key(self.internal_storage.prefix, self.executor_id, callgroup_id)
        self.internal_storage.put_func(func_key, func_module_str)
        host_job_meta['func_upload_time'] = time.time() - func_upload_time
        host_job_meta['func_upload_timestamp'] = time.time()

        if not self.log_level:
            func_and_data_size = wrenutil.sizeof_fmt(host_job_meta['func_module_bytes']+host_job_meta['data_size_bytes'])
            log_msg = '- Total: {}'.format(func_and_data_size)
            print(log_msg)

        def invoke(data_str, executor_id, callgroup_id, call_id, func_key,
                   host_job_meta, agg_data_key=None, data_byte_range=None):
            data_key, output_key, status_key = create_keys(self.internal_storage.prefix,
                                                           executor_id, callgroup_id, call_id)
            host_job_meta['job_invoke_timestamp'] = time.time()

            if agg_data_key is None:
                data_upload_time = time.time()
                self.internal_storage.put_data(data_key, data_str)
                data_upload_time = time.time() - data_upload_time
                host_job_meta['data_upload_time'] = data_upload_time
                host_job_meta['data_upload_timestamp'] = time.time()

                data_key = data_key
            else:
                data_key = agg_data_key

            return self.invoke_with_keys(func_key, data_key,
                                         output_key, status_key,
                                         executor_id, callgroup_id,
                                         call_id, extra_env,
                                         extra_meta, data_byte_range,
                                         host_job_meta.copy(),
                                         job_max_runtime,
                                         overwrite_invoke_args=overwrite_invoke_args)

        N = len(data)
        call_result_objs = []

        log_msg = 'Executor ID {} Starting function invocation: {}() - Total: {} activations'.format(self.executor_id, func_name, N)
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg)

        for i in range(N):
            call_id = "{:05d}".format(i)

            data_byte_range = None
            if agg_data_key is not None:
                data_byte_range = agg_data_ranges[i]

            cb = pool.apply_async(invoke, (data_strs[i], self.executor_id,
                                           callgroup_id, call_id, func_key,
                                           host_job_meta.copy(),
                                           agg_data_key,
                                           data_byte_range))

            call_result_objs.append(cb)

        res = [c.get() for c in call_result_objs]
        pool.close()
        pool.join()

        return res

    def reduce(self, reduce_function, list_of_futures, parts_per_object,
               reducer_one_per_object, extra_env, extra_meta):
        """
        Apply a function across all futures.
        """
        executor_id = self.executor_id
        map_iterdata = [[list_of_futures, ]]

        if parts_per_object and reducer_one_per_object:
            prev_total_partitons = 0
            map_iterdata = []
            for total_partitions in parts_per_object:
                map_iterdata.append([list_of_futures[prev_total_partitons:prev_total_partitons+total_partitions]])
                prev_total_partitons = prev_total_partitons + total_partitions

        def reduce_function_wrapper(fut_list, internal_storage, storage, ibm_cos):
            logger.info('Waiting for results')
            if 'SHOW_MEMORY_USAGE' in os.environ:
                show_memory = eval(os.environ['SHOW_MEMORY_USAGE'])
            else:
                show_memory = False
            # Wait for all results
            wait(fut_list, executor_id, internal_storage, download_results=True)
            results = [f.result() for f in fut_list if f.done and not f.futures]
            reduce_func_args = {'results': results}

            if show_memory:
                logger.debug("Memory usage after getting the results: {}".format(wrenutil.get_current_memory_usage()))

            # Run reduce function
            func_sig = inspect.signature(reduce_function)
            if 'storage' in func_sig.parameters:
                reduce_func_args['storage'] = storage
            if 'ibm_cos' in func_sig.parameters:
                reduce_func_args['ibm_cos'] = ibm_cos

            return reduce_function(**reduce_func_args)
            #result = reduce_function(**reduce_func_args)
            #run_statuses = [f.run_status for f in fut_list]
            #invoke_statuses = [f.invoke_status for f in fut_list]

            #return {'fn_result': result, 'run_statuses': run_statuses, 'invoke_statuses': invoke_statuses}

        return self._map(reduce_function_wrapper, map_iterdata, extra_env=extra_env,
                         extra_meta=extra_meta, original_func_name=reduce_function.__name__)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging.config


def default_config(log_level='INFO'):
    if log_level == 'DEBUG_BOTO3':
        log_level = 'DEBUG'
        logging.getLogger('ibm_boto3').setLevel(logging.DEBUG)
        logging.getLogger('ibm_botocore').setLevel(logging.DEBUG)

    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'standard'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': True
            }
        }
    })


def ow_config(log_level='INFO'):
    if log_level == 'DEBUG_BOTO3':
        log_level = 'DEBUG'
        logging.getLogger('ibm_boto3').setLevel(logging.DEBUG)
        logging.getLogger('ibm_botocore').setLevel(logging.DEBUG)

    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '[%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
                'stream': 'ext://sys.stdout'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': True
            }
        }
    })



import sys
import yaml
import os
import unittest
import ibm_boto3
from ibm_botocore.client import Config
from ibm_botocore.client import ClientError
import pywren_ibm_cloud as pywren
import urllib.request
import logging

PREFIX = '__pywren.test'

#logging.basicConfig(level=logging.DEBUG)

try:
    dir_path = os.path.dirname(__file__)
    path = os.path.join(dir_path, 'data')
    with open(path, 'r') as data_file:
        TEST_FILES_URLS = [url for url in data_file.read().split()]
except:
    print("can't open data file")
    sys.exit()

try:
    if 'PYWREN_CONFIG_FILE' in os.environ:
        config_path = os.environ['PYWREN_CONFIG_FILE']
    else:
        config_path = os.path.join(os.path.expanduser("~/.pywren_config"))
    with open(config_path, 'r') as config_file:
        CONFIG = yaml.safe_load(config_file)
except:
    print("can't open config file")
    sys.exit()


def initCos():
    return ibm_boto3.resource("s3",
                              ibm_api_key_id=CONFIG['ibm_cos']['api_key'],
                              ibm_auth_endpoint='https://iam.ng.bluemix.net/oidc/token',
                              config=Config(signature_version="oauth"),
                              endpoint_url=CONFIG['ibm_cos']['endpoint']
                              )


def putFileToCOS(cos, bucket_name, key, bytes):
    try:
        cos.Object(bucket_name, key).put(Body=bytes)
        print("Upload file: {} - SUCCESS".format(key))
    except ClientError as be:
        print("CLIENT ERROR: {0}\n".format(be))
    except Exception as e:
        print("Unable to create bucket: {0}".format(e))


def getFilenamesFromCOS(cos, bucket_name, prefix):
    print("Retrieving items' names from bucket: {0}, prefix: {1}".format(bucket_name, prefix))
    result = []
    try:
        for data in cos.Bucket(bucket_name).objects.filter(Prefix=prefix):
            result.append(data.key)
    except ClientError as be:
        print("CLIENT ERROR: {0}\n".format(be))
    except Exception as e:
        print("Unable to delete item: {0}".format(e))
    return result


def getFileFromCOS(cos, bucket_name, key):
    print("Retrieving item from bucket: {0}, key: {1}".format(bucket_name, key))
    try:
        file = cos.Object(bucket_name, key).get()
        return file["Body"].read()
    except ClientError as be:
        print("CLIENT ERROR: {0}\n".format(be))
    except Exception as e:
        print("Unable to retrieve file contents: {0}".format(e))


def deleteFileFromCOS(cos, bucket_name, key):
    try:
        cos.Object(bucket_name, key).delete()
        print("File: {0} deleted!".format(key))
    except ClientError as be:
        print("CLIENT ERROR: {0}\n".format(be))
    except Exception as e:
        print("Unable to delete item: {0}".format(e))


class TestPywren(unittest.TestCase):

    def hello_world(self, param):
        return "Hello World!"

    def simple_map_function(self, x, y):
        return x + y

    def simple_reduce_function(self, results):
        total = 0
        for map_result in results:
            total = total + map_result
        return total

    def test_call_async(self):
        pw = pywren.ibm_cf_executor()
        pw.call_async(self.hello_world, "")
        result = pw.get_result()
        self.assertEqual(result, "Hello World!")

        pw = pywren.ibm_cf_executor()
        pw.call_async(self.simple_map_function, [4, 6])
        result = pw.get_result()
        self.assertEqual(result, 10)

        pw = pywren.ibm_cf_executor()
        pw.call_async(self.simple_map_function, {'x': 2, 'y': 8})
        result = pw.get_result()
        self.assertEqual(result, 10)

    def test_map(self):
        iterdata = [[1, 1], [2, 2], [3, 3], [4, 4]]
        pw = pywren.ibm_cf_executor()
        pw.map(self.simple_map_function, iterdata)
        result = pw.get_result()
        self.assertEqual(result, [2, 4, 6, 8])

    def test_map_reduce(self):
        iterdata = [[1, 1], [2, 2], [3, 3], [4, 4]]
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.simple_map_function, iterdata, self.simple_reduce_function)
        result = pw.get_result()
        self.assertEqual(result, 20)

    def test_multiple_executions(self):
        pw = pywren.ibm_cf_executor()
        iterdata = [[1, 1], [2, 2]]
        pw.map(self.simple_map_function, iterdata)
        iterdata = [[3, 3], [4, 4]]
        pw.map(self.simple_map_function, iterdata)
        result = pw.get_result()
        self.assertEqual(result, [2, 4, 6, 8])

        iterdata = [[1, 1], [2, 2]]
        futures1 = pw.map(self.simple_map_function, iterdata)
        result1 = pw.get_result(futures=futures1)
        iterdata = [[3, 3], [4, 4]]
        futures2 = pw.map(self.simple_map_function, iterdata)
        result2 = pw.get_result(futures=futures2)
        self.assertEqual(result1, [2, 4])
        self.assertEqual(result2, [6, 8])


def initTests():
    print('Uploading test files...')

    cos = initCos()
    result_to_compare = 1  # including result's word
    i = 0
    for url in TEST_FILES_URLS:
        content = urllib.request.urlopen(url).read()
        putFileToCOS(cos, CONFIG['pywren']['storage_bucket'], PREFIX + '/test' + str(i), content)
        result_to_compare += len(content.split())
        i += 1

    putFileToCOS(cos, CONFIG['pywren']['storage_bucket'], PREFIX + '/result', str(result_to_compare).encode())

    print("ALL DONE")


def cleanTests():
    print('Deleting test files...')

    cos = initCos()
    for key in getFilenamesFromCOS(cos, CONFIG['pywren']['storage_bucket'], PREFIX):
        deleteFileFromCOS(cos, CONFIG['pywren']['storage_bucket'], key)

    print("ALL DONE")


class TestPywrenCos(unittest.TestCase):

    def my_map_function_bucket(self, bucket, key, data_stream, ibm_cos):
        print('I am processing the object {}'.format(key))
        counter = {}

        data = data_stream.read()
        temp = ibm_cos.get_object(Bucket=bucket, Key=key)['Body'].read()

        for line in data.splitlines():
            for word in line.decode('utf-8').split():
                if word not in counter:
                    counter[word] = 1
                else:
                    counter[word] += 1

        return counter

    def my_map_function_key(self, key, data_stream, ibm_cos):
        print('I am processing the object {}'.format(key))
        counter = {}

        data = data_stream.read()
        temp = ibm_cos.get_object(Bucket=key.split('/')[0], Key='/'.join(key.split('/')[1:]))['Body'].read()

        for line in data.splitlines():
            for word in line.decode('utf-8').split():
                if word not in counter:
                    counter[word] = 1
                else:
                    counter[word] += 1

        return counter

    def my_map_function_url(self, url, data_stream):
        print('I am processing the object from {}'.format(url))
        counter = {}

        data = data_stream.read()

        for line in data.splitlines():
            for word in line.decode('utf-8').split():
                if word not in counter:
                    counter[word] = 1
                else:
                    counter[word] += 1

        return counter

    def my_map_function_storage_handler(self, key_i, ibm_cos):
        print('I am processing the object {}'.format(key_i))
        counter = {}

        data = ibm_cos.get_object(Bucket=CONFIG['pywren']['storage_bucket'], Key=key_i)['Body'].read()

        for line in data.splitlines():
            for word in line.decode('utf-8').split():
                if word not in counter:
                    counter[word] = 1
                else:
                    counter[word] += 1

        return counter

    def my_reduce_function(self, results):
        final_result = 0

        for count in results:
            for word in count:
                final_result += count[word]

        return final_result

    def checkResult(self, cos, result):
        result_to_compare = getFileFromCOS(cos, CONFIG['pywren']['storage_bucket'], PREFIX + '/result')

        if isinstance(result, list):
            total = 0
            for r in result:
                total += r
        else:
            total = result

        self.assertEqual(total, int(result_to_compare))

    def test_map_reduce_cos_bucket(self):
        data_prefix = CONFIG['pywren']['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_bucket, data_prefix, self.my_reduce_function)
        result = pw.get_result()
        self.checkResult(initCos(), result)

    def test_map_reduce_cos_bucket_one_reducer_per_object(self):
        data_prefix = CONFIG['pywren']['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_bucket, data_prefix, self.my_reduce_function, reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(initCos(), result)

    def test_map_reduce_cos_key(self):
        cos = initCos()
        bucket_name = CONFIG['pywren']['storage_bucket']
        iterdata = [bucket_name + '/' + key for key in getFilenamesFromCOS(cos, bucket_name, PREFIX)]
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_key, iterdata, self.my_reduce_function)
        result = pw.get_result()
        self.checkResult(cos, result)

    def test_map_reduce_cos_key_one_reducer_per_object(self):
        cos = initCos()
        bucket_name = CONFIG['pywren']['storage_bucket']
        iterdata = [bucket_name + '/' + key for key in getFilenamesFromCOS(cos, bucket_name, PREFIX)]
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_key, iterdata, self.my_reduce_function, reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(cos, result)

    def test_map_reduce_url(self):
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_url, TEST_FILES_URLS, self.my_reduce_function)
        result = pw.get_result()
        self.checkResult(initCos(), result + 1)

    def test_storage_handler(self):
        cos = initCos()
        bucket_name = CONFIG['pywren']['storage_bucket']
        iterdata = [key for key in getFilenamesFromCOS(cos, bucket_name, PREFIX)]
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_storage_handler, iterdata, self.my_reduce_function)
        result = pw.get_result()
        self.checkResult(cos, result)

    def test_chunks_bucket(self):
        data_prefix = CONFIG['pywren']['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_bucket, data_prefix, self.my_reduce_function, chunk_size=1 * 1024 ** 2)
        result = pw.get_result()
        self.checkResult(initCos(), result)

    def test_chunks_bucket_one_reducer_per_object(self):
        data_prefix = CONFIG['pywren']['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor()
        pw.map_reduce(self.my_map_function_bucket, data_prefix, self.my_reduce_function, chunk_size=1 * 1024 ** 2,
                      reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(initCos(), result)


if __name__ == '__main__':
    if len(sys.argv) <= 1:
        task = 'full'
    else:
        task = sys.argv[1]

    if task == 'init':
        initTests()
    elif task == 'clean':
        cleanTests()
    else:
        suite = unittest.TestSuite()
        if task == 'pywren':
            suite.addTest(unittest.makeSuite(TestPywren))
        elif task == 'pywren_cos':
            suite.addTest(unittest.makeSuite(TestPywrenCos))
        elif task == 'full':
            suite.addTest(unittest.makeSuite(TestPywren))
            suite.addTest(unittest.makeSuite(TestPywrenCos))
        elif task == 'test_call_async':
            suite.addTest(TestPywren('test_call_async'))
        elif task == 'test_map':
            suite.addTest(TestPywren('test_map'))
        elif task == 'test_map_reduce':
            suite.addTest(TestPywren('test_map_reduce'))
        elif task == 'test_multiple_executions':
            suite.addTest(TestPywren('test_multiple_executions'))
        elif task == 'test_map_reduce_cos_bucket':
            suite.addTest(TestPywrenCos('test_map_reduce_cos_bucket'))
        elif task == 'test_map_reduce_cos_bucket_one_reducer_per_object':
            suite.addTest(TestPywrenCos('test_map_reduce_cos_bucket_one_reducer_per_object'))
        elif task == 'test_map_reduce_cos_key':
            suite.addTest(TestPywrenCos('test_map_reduce_cos_key'))
        elif task == 'test_map_reduce_cos_key_one_reducer_per_object':
            suite.addTest(TestPywrenCos('test_map_reduce_cos_key_one_reducer_per_object'))
        elif task == 'test_map_reduce_url':
            suite.addTest(TestPywrenCos('test_map_reduce_url'))
        elif task == 'test_storage_handler':
            suite.addTest(TestPywrenCos('test_storage_handler'))
        elif task == 'test_chunks_bucket':
            suite.addTest(TestPywrenCos('test_chunks_bucket'))
        elif task == 'test_chunks_bucket_one_reducer_per_object':
            suite.addTest(TestPywrenCos('test_chunks_bucket_one_reducer_per_object'))
        else:
            print('Unknown Command... use: "init", "pywren", "pywren_cos", "clean" or a test function name.')
            sys.exit()

        runner = unittest.TextTestRunner()
        runner.run(suite)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import json
from pywren_ibm_cloud.utils import version_str, is_cf_cluster

STORAGE_BACKEND_DEFAULT = 'ibm_cos'
COS_BUCKET_DEFAULT = "pywren.data"
COS_PREFIX_DEFAULT = "pywren.jobs"
COS_AUTH_ENDPOINT_DEFAULT = 'https://iam.cloud.ibm.com'

RUNTIME_DEFAULT_35 = 'ibmfunctions/pywren:3.5'
RUNTIME_DEFAULT_36 = 'ibmfunctions/action-python-v3.6'
RUNTIME_DEFAULT_37 = 'ibmfunctions/action-python-v3.7'

RUNTIME_TIMEOUT_DEFAULT = 600000  # Default: 600000 milliseconds => 10 minutes
RUNTIME_MEMORY_DEFAULT = 256  # Default: 256 MB
RUNTIME_TIMEOUT = 600  # Default: 600 seconds => 10 minutes

DATA_CLEANER_DEFAULT = False
MAX_AGG_DATA_SIZE = 4e6
INVOCATION_RETRY_DEFAULT = True
RETRY_SLEEPS_DEFAULT = [1, 5, 10, 15, 20]
RETRIES_DEFAULT = 5
AMQP_URL_DEFAULT = None


def load(config_filename):
    import yaml
    with open(config_filename, 'r') as config_file:
        res = yaml.safe_load(config_file)

    if 'pywren' not in res:
        raise Exception("pywren section mandatory in {}".format(config_filename))

    if 'pywren' in res and res['pywren']['storage_bucket'] == '<BUCKET_NAME>':
        raise Exception(
            "{} has bucket name as {} -- make sure you change the default container".format(
                config_filename, res['pywren']['storage_bucket']))

    if 'ibm_cf' not in res:
        raise Exception("ibm_cf section mandatory in {}".format(config_filename))

    if 'endpoint' in res['ibm_cf'] and res['ibm_cf']['endpoint'] == '<CF_API_ENDPOINT>':
        raise Exception(
            "{} has CF API endpoint as {} -- make sure you change the default CF API endpoint".format(
                config_filename, res['ibm_cf']['endpoint']))
    if 'namespace' in res['ibm_cf'] and res['ibm_cf']['namespace'] == '<CF_NAMESPACE>':
        raise Exception(
            "{} has namespace as {} -- make sure you change the default namespace".format(
                config_filename, res['ibm_cf']['namespace']))
    if 'api_key' in res['ibm_cf'] and res['ibm_cf']['api_key'] == '<CF_API_KEY>':
        raise Exception(
            "{} has CF API key as {} -- make sure you change the default CF API key".format(
                config_filename, res['ibm_cf']['api_key']))

    if 'ibm_cos' in res:
        if 'endpoint' in res['ibm_cos'] and res['ibm_cos']['endpoint'] == '<COS_API_ENDPOINT>':
            raise Exception(
                "{} has CF API endpoint as {} -- make sure you change the default COS API endpoint".format(
                    config_filename, res['ibm_cos']['endpoint']))
        if 'api_key' in res['ibm_cos'] and res['ibm_cos']['api_key'] == '<COS_API_KEY>':
            raise Exception(
                "{} has CF API key as {} -- make sure you change the default COS API key".format(
                    config_filename, res['ibm_cos']['api_key']))

    return res


def get_default_home_filename():
    default_home_filename = os.path.join(os.path.expanduser("~/.pywren_config"))
    return default_home_filename


def get_default_config_filename():
    """
    First checks .pywren_config
    then checks PYWREN_CONFIG_FILE environment variable
    then ~/.pywren_config
    """
    if 'PYWREN_CONFIG_FILE' in os.environ:
        config_filename = os.environ['PYWREN_CONFIG_FILE']
        # FIXME log this

    elif os.path.exists(".pywren_config"):
        config_filename = os.path.abspath('.pywren_config')

    else:
        config_filename = get_default_home_filename()

    return config_filename


def default(config_data=None):
    """
    First checks .pywren_config
    then checks PYWREN_CONFIG_FILE environment variable
    then ~/.pywren_config
    """
    if not config_data:
        if 'PYWREN_CONFIG' in os.environ:
            config_data = json.loads(os.environ.get('PYWREN_CONFIG'))
        else:
            config_filename = get_default_config_filename()
            if config_filename is None:
                raise ValueError("could not find configuration file")

            config_data = load(config_filename)

    if not set(('pywren', 'ibm_cf', 'ibm_cos')).issubset(set(config_data)):
        raise Exception("pywren, ibm_cf and ibm_cos sections are mandatory in config")

    if 'storage_backend' not in config_data['pywren']:
        config_data['pywren']['storage_backend'] = STORAGE_BACKEND_DEFAULT
    if 'storage_bucket' not in config_data['pywren']:
        config_data['pywren']['storage_bucket'] = COS_BUCKET_DEFAULT
    if 'storage_prefix' not in config_data['pywren']:
        config_data['pywren']['storage_prefix'] = COS_PREFIX_DEFAULT
    if 'data_cleaner' not in config_data['pywren']:
        config_data['pywren']['data_cleaner'] = DATA_CLEANER_DEFAULT
    if 'invocation_retry' not in config_data['pywren']:
        config_data['pywren']['invocation_retry'] = INVOCATION_RETRY_DEFAULT
    if 'retry_sleeps' not in config_data['pywren']:
        config_data['pywren']['retry_sleeps'] = RETRY_SLEEPS_DEFAULT
    if 'retries' not in config_data['pywren']:
        config_data['pywren']['retries'] = RETRIES_DEFAULT
    if 'runtime_memory' not in config_data['pywren']:
        config_data['pywren']['runtime_memory'] = RUNTIME_MEMORY_DEFAULT
    if 'runtime_timeout' not in config_data['pywren']:
        config_data['pywren']['runtime_timeout'] = RUNTIME_TIMEOUT_DEFAULT
    if 'runtime' not in config_data['pywren']:
        this_version_str = version_str(sys.version_info)
        if this_version_str == '3.5':
            config_data['pywren']['runtime'] = RUNTIME_DEFAULT_35
        elif this_version_str == '3.6':
            config_data['pywren']['runtime'] = RUNTIME_DEFAULT_36
        elif this_version_str == '3.7':
            config_data['pywren']['runtime'] = RUNTIME_DEFAULT_37

    if 'ibm_cos' in config_data and 'ibm_auth_endpoint' not in config_data['ibm_cos']:
        config_data['ibm_cos']['ibm_auth_endpoint'] = COS_AUTH_ENDPOINT_DEFAULT

    # True or False depending on whether this code is executed within CF cluster or not
    config_data['ibm_cf']['is_cf_cluster'] = is_cf_cluster()

    if 'rabbitmq' not in config_data or not config_data['rabbitmq'] \
       or 'amqp_url' not in config_data['rabbitmq']:
        config_data['rabbitmq'] = {}
        config_data['rabbitmq']['amqp_url'] = None

    return config_data


def extract_storage_config(config):
    storage_config = dict()

    storage_config['storage_backend'] = config['pywren']['storage_backend']
    storage_config['storage_prefix'] = config['pywren']['storage_prefix']
    storage_config['storage_bucket'] = config['pywren']['storage_bucket']

    if 'ibm_cos' in config:
        required_parameters_1 = ('endpoint', 'api_key')
        required_parameters_2 = ('endpoint', 'secret_key', 'access_key')

        if set(required_parameters_1) <= set(config['ibm_cos']) or \
                set(required_parameters_2) <= set(config['ibm_cos']):
            storage_config['ibm_cos'] = config['ibm_cos']
        else:
            raise Exception('You must provide {} or {} to access to IBM COS'.format(required_parameters_1,
                                                                                    required_parameters_2))

    if 'swift' in config:
        required_parameters = ('auth_url', 'user_id', 'project_id', 'password', 'region')

        if set(required_parameters) <= set(config['swift']):
            storage_config['swift'] = config['swift']
        else:
            raise Exception('You must provide {} to access to Swift'.format(required_parameters))

    return storage_config


def extract_cf_config(config):
    cf_config = config['ibm_cf']
    cf_config['runtime'] = config['pywren']['runtime']
    cf_config['runtime_timeout'] = int(config['pywren']['runtime_timeout'])
    cf_config['runtime_memory'] = int(config['pywren']['runtime_memory'])

    return cf_config



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import json
import logging
import os
import subprocess
import time
import traceback
import pika
import multiprocessing
from pywren_ibm_cloud import version
from pywren_ibm_cloud import wrenconfig
from pywren_ibm_cloud import wrenlogging
from pywren_ibm_cloud.storage import storage
from pywren_ibm_cloud.utils import sizeof_fmt
from pywren_ibm_cloud.action.jobrunner import jobrunner


logging.getLogger('pika').setLevel(logging.CRITICAL)
logger = logging.getLogger('wrenhandler')

JOBRUNNER_PATH = "pywren_ibm_cloud/action/jobrunner.py"

PYTHON_MODULE_PATH = "/tmp/pymodules"
JOBRUNNER_CONFIG_FILENAME = "/tmp/jobrunner.config.json"
JOBRUNNER_STATS_FILENAME = "/tmp/jobrunner.stats.txt"
PYWREN_LIBS_PATH = '/action/pywren_ibm_cloud/libs'


def free_disk_space(dirname):
    """
    Returns the number of free bytes on the mount point containing DIRNAME
    """
    s = os.statvfs(dirname)
    return s.f_bsize * s.f_bavail


def b64str_to_bytes(str_data):
    str_ascii = str_data.encode('ascii')
    byte_data = base64.b64decode(str_ascii)
    return byte_data


def get_server_info():
    server_info = {'uname': subprocess.check_output("uname -a", shell=True).decode("ascii").strip(),
                   'ip_adress': subprocess.check_output("hostname -I", shell=True).decode("ascii").strip()}
    """
    if os.path.exists("/proc"):
        server_info.update({'/proc/cpuinfo': open("/proc/cpuinfo", 'r').read(),
                            '/proc/meminfo': open("/proc/meminfo", 'r').read(),
                            '/proc/self/cgroup': open("/proc/meminfo", 'r').read(),
                            '/proc/cgroups': open("/proc/cgroups", 'r').read()})
    """
    return server_info


def ibm_cloud_function_handler(event):
    start_time = time.time()
    logger.info("Starting handler")
    response_status = {'exception': None}
    response_status['start_time'] = start_time

    context_dict = {
        'ibm_cf_request_id': os.environ.get("__OW_ACTIVATION_ID"),
        'ibm_cf_hostname': os.environ.get("HOSTNAME"),
        'ibm_cf_python_version': os.environ.get("PYTHON_VERSION"),
    }

    config = event['config']
    storage_config = wrenconfig.extract_storage_config(config)

    log_level = event['log_level']
    wrenlogging.ow_config(log_level)

    call_id = event['call_id']
    callgroup_id = event['callgroup_id']
    executor_id = event['executor_id']
    job_max_runtime = event.get("job_max_runtime", 590)  # default for CF
    status_key = event['status_key']
    func_key = event['func_key']
    data_key = event['data_key']
    data_byte_range = event['data_byte_range']
    output_key = event['output_key']
    extra_env = event.get('extra_env', {})

    response_status['call_id'] = call_id
    response_status['callgroup_id'] = callgroup_id
    response_status['executor_id'] = executor_id
    response_status['func_key'] = func_key
    response_status['data_key'] = data_key
    response_status['output_key'] = output_key
    response_status['status_key'] = status_key

    try:
        if version.__version__ != event['pywren_version']:
            raise Exception("WRONGVERSION", "PyWren version mismatch",
                            version.__version__, event['pywren_version'])

        # response_status['free_disk_bytes'] = free_disk_space("/tmp")

        custom_env = {'PYWREN_CONFIG': json.dumps(config),
                      'STORAGE_CONFIG': json.dumps(storage_config),
                      'PYWREN_EXECUTOR_ID':  executor_id,
                      'PYTHONPATH': "{}:{}".format(os.getcwd(), PYWREN_LIBS_PATH),
                      'PYTHONUNBUFFERED': 'True'}

        os.environ.update(custom_env)
        os.environ.update(extra_env)

        # pass a full json blob
        jobrunner_config = {'func_key': func_key,
                            'data_key': data_key,
                            'log_level': log_level,
                            'data_byte_range': data_byte_range,
                            'python_module_path': PYTHON_MODULE_PATH,
                            'output_key': output_key,
                            'stats_filename': JOBRUNNER_STATS_FILENAME}

        with open(JOBRUNNER_CONFIG_FILENAME, 'w') as jobrunner_fid:
            json.dump(jobrunner_config, jobrunner_fid)

        if os.path.exists(JOBRUNNER_STATS_FILENAME):
            os.remove(JOBRUNNER_STATS_FILENAME)

        setup_time = time.time()
        response_status['setup_time'] = setup_time - start_time

        result_queue = multiprocessing.Queue()
        jr = jobrunner(JOBRUNNER_CONFIG_FILENAME, result_queue)
        jr.daemon = True
        logger.info("Launched jobrunner process")
        jr.start()
        jr.join(job_max_runtime)
        response_status['exec_time'] = time.time() - setup_time

        if jr.is_alive():
            # If process is still alive after jr.join(job_max_runtime), kill it
            logger.error("Process exceeded maximum runtime of {} sec".format(job_max_runtime))
            # Send the signal to all the process groups
            jr.terminate()
            raise Exception("OUTATIME",  "Process executed for too long and was killed")

        try:
            # Only 1 message is returned by jobrunner
            result_queue.get(block=False)
        except Exception:
            # If no mesage, this means that the process was killed due memory usage
            raise Exception("OUTOFMEMORY",  "Process exceeded maximum memory and was killed")

        # print(subprocess.check_output("find {}".format(PYTHON_MODULE_PATH), shell=True))
        # print(subprocess.check_output("find {}".format(os.getcwd()), shell=True))

        if os.path.exists(JOBRUNNER_STATS_FILENAME):
            with open(JOBRUNNER_STATS_FILENAME, 'r') as fid:
                for l in fid.readlines():
                    key, value = l.strip().split(" ")
                    try:
                        response_status[key] = float(value)
                    except Exception:
                        response_status[key] = value

        response_status['host_submit_time'] = event['host_submit_time']
        # response_status['server_info'] = get_server_info()
        response_status.update(context_dict)
        response_status['end_time'] = time.time()

    except Exception as e:
        # internal runtime exceptions
        logger.error("There was an exception: {}".format(str(e)))
        response_status['end_time'] = time.time()
        response_status['exception'] = str(e)
        response_status['exception_args'] = e.args
        response_status['exception_traceback'] = traceback.format_exc()

    finally:
        rabbit_amqp_url = config['rabbitmq'].get('amqp_url')
        if rabbit_amqp_url:
            params = pika.URLParameters(rabbit_amqp_url)
            connection = pika.BlockingConnection(params)
            channel = connection.channel()
            channel.queue_declare(queue=executor_id, auto_delete=True)
            status = 'ok'
            if response_status['exception']:
                status = 'error'
            try:
                new_futures = response_status['new_futures']
                channel.basic_publish(exchange='', routing_key=executor_id,
                                      body='{}/{}:{}:{}'.format(callgroup_id, call_id,
                                                                status,  new_futures))
                logger.info("Status sent to rabbitmq")
            except Exception:
                logger.error("Unable to send status to rabbitmq")

            connection.close()

        store_status = True
        if 'STORE_STATUS' in extra_env:
            store_status = eval(extra_env['STORE_STATUS'])

        if store_status:
            internal_storage = storage.InternalStorage(storage_config)
            logger.info("Storing {} - Size: {}".format(status_key, sizeof_fmt(len(response_status))))
            internal_storage.put_data(status_key, json.dumps(response_status))



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from pywren_ibm_cloud.storage.exceptions import StorageConfigMismatchError

func_key_suffix = "func.pickle"
agg_data_key_suffix = "aggdata.pickle"
data_key_suffix = "data.pickle"
output_key_suffix = "output.pickle"
status_key_suffix = "status.json"


def create_func_key(prefix, executor_id, callgroup_id):
    """
    Create function key
    :param prefix: prefix
    :param executor_id: callset's ID
    :return: function key
    """
    func_key = os.path.join(prefix, executor_id, callgroup_id, func_key_suffix)
    return func_key


def create_agg_data_key(prefix, executor_id, callgroup_id):
    """
    Create aggregate data key
    :param prefix: prefix
    :param executor_id: callset's ID
    :return: a key for aggregate data
    """
    agg_data_key = os.path.join(prefix, executor_id, callgroup_id, agg_data_key_suffix)
    return agg_data_key


def create_data_key(prefix, executor_id, callgroup_id, call_id):
    """
    Create data key
    :param prefix: prefix
    :param executor_id: callset's ID
    :param call_id: call's ID
    :return: data key
    """
    return os.path.join(prefix, executor_id, callgroup_id, call_id, data_key_suffix)


def create_output_key(prefix, executor_id, callgroup_id, call_id):
    """
    Create output key
    :param prefix: prefix
    :param executor_id: callset's ID
    :param call_id: call's ID
    :return: output key
    """
    return os.path.join(prefix, executor_id, callgroup_id, call_id, output_key_suffix)


def create_status_key(prefix, executor_id, callgroup_id, call_id):
    """
    Create status key
    :param prefix: prefix
    :param executor_id: callset's ID
    :param call_id: call's ID
    :return: status key
    """
    return os.path.join(prefix, executor_id, callgroup_id, call_id, status_key_suffix)


def create_keys(prefix, executor_id, callgroup_id, call_id):
    """
    Create keys for data, output and status given callset and call IDs.
    :param prefix: prefix
    :param executor_id: callset's ID
    :param call_id: call's ID
    :return: data_key, output_key, status_key
    """
    data_key = create_data_key(prefix, executor_id, callgroup_id, call_id)
    output_key = create_output_key(prefix, executor_id, callgroup_id, call_id)
    status_key = create_status_key(prefix, executor_id, callgroup_id, call_id)
    return data_key, output_key, status_key


def get_storage_path(storage_config):
    storage_bucket = storage_config['storage_bucket']
    storage_prefix = storage_config['storage_prefix']
    storage_backend = storage_config['storage_backend']

    if storage_backend != 'ibm_cos' and storage_backend != 'swift':
        raise NotImplementedError(
            ("Using {} as storage backend is not supported yet").format(
             storage_backend))

    return [storage_backend, storage_bucket, storage_prefix]


def check_storage_path(config, prev_path):
    current_path = get_storage_path(config)
    if current_path != prev_path:
        raise StorageConfigMismatchError(current_path, prev_path)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import time
import enum
import json
import signal
import logging
import pywren_ibm_cloud.invokers as invokers
import pywren_ibm_cloud.wrenconfig as wrenconfig
from pywren_ibm_cloud import wrenlogging
from pywren_ibm_cloud.storage import storage
from pywren_ibm_cloud.executor import Executor
from pywren_ibm_cloud.wait import wait, ALL_COMPLETED
from pywren_ibm_cloud.utils import timeout_handler, is_notebook
from pywren_ibm_cloud.storage.cleaner import clean_os_bucket

logger = logging.getLogger(__name__)


class ExecutorState(enum.Enum):
    new = 1
    running = 2
    ready = 3
    result = 4
    finished = 5
    error = 6


class ibm_cf_executor:

    def __init__(self, config=None, runtime=None, runtime_memory=None, log_level=None, rabbitmq_monitor=False):
        """
        Initialize and return an executor class.

        :param config: Settings passed in here will override those in `pywren_config`. Default None.
        :param runtime: Runtime name to use. Default None.
        :param runtime_memory: memory to use in the runtime
        :param log_level: log level to use during the execution
        :param rabbitmq_monitor: use rabbitmq as monitoring system
        :return `executor` object.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
        """
        self.start_time = time.time()
        self._state = ExecutorState.new

        if config is None:
            self.config = wrenconfig.default()
        else:
            self.config = wrenconfig.default(config)

        self.is_cf_cluster = self.config['ibm_cf']['is_cf_cluster']
        self.data_cleaner = self.config['pywren']['data_cleaner']

        # Overwrite runtime variables
        if runtime:
            self.config['pywren']['runtime'] = runtime
        if runtime_memory:
            self.config['pywren']['runtime_memory'] = int(runtime_memory)

        # Log level Configuration
        self.log_level = log_level
        if not self.log_level:
            if(logger.getEffectiveLevel() != logging.WARNING):
                self.log_level = logging.getLevelName(logger.getEffectiveLevel())
        if self.log_level:
            os.environ["PYWREN_LOG_LEVEL"] = self.log_level
            if not self.is_cf_cluster:
                wrenlogging.default_config(self.log_level)

        # RabbitMQ monitor configuration
        self.rabbitmq_monitor = rabbitmq_monitor
        if self.rabbitmq_monitor:
            os.environ["PYWREN_RABBITMQ_MONITOR"] = 'True'
        if not self.rabbitmq_monitor:
            self.config['rabbitmq']['amqp_url'] = None

        storage_config = wrenconfig.extract_storage_config(self.config)
        self.internal_storage = storage.InternalStorage(storage_config)

        invoker = invokers.IBMCloudFunctionsInvoker(self.config)
        self.executor = Executor(invoker, self.config, self.internal_storage)
        self.executor_id = self.executor.executor_id

        self.futures = []

    def call_async(self, func, data, extra_env=None, extra_meta=None, timeout=wrenconfig.RUNTIME_TIMEOUT):
        """
        For run one function execution
        :param func: the function to map over the data
        :param data: input data
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> future = pw.call_async(foo, data)
        """
        if self._state == ExecutorState.finished or self._state == ExecutorState.error:
            raise Exception('You cannot run pw.call_async() in the current state,'
                            ' create a new pywren.ibm_cf_executor() instance.')

        future = self.executor.call_async(func, data, extra_env, extra_meta, timeout)[0]
        self.futures.append(future)
        self._state = ExecutorState.running

        return future

    def map(self, map_function, map_iterdata, extra_env=None, extra_meta=None,
            chunk_size=None, remote_invocation=False, timeout=wrenconfig.RUNTIME_TIMEOUT,
            remote_invocation_groups=100, invoke_pool_threads=128,
            data_all_as_one=True, overwrite_invoke_args=None, exclude_modules=None):
        """
        :param func: the function to map over the data
        :param iterdata: An iterable of input data
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.
        :param chunk_size: the size of the data chunks. 'None' for processing the whole file in one map
        :param data_type: the type of the data. Now allowed: None (files with newline) and csv.
        :param invoke_pool_threads: Number of threads to use to invoke.
        :param data_all_as_one: upload the data as a single object. Default True
        :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
        :param exclude_modules: Explicitly keep these modules from pickled dependencies.
        :return: A list with size `len(iterdata)` of futures for each job
        :rtype: list of futures.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> futures = pw.map(foo, data_list)
        """
        if self._state == ExecutorState.finished or self._state == ExecutorState.error:
            raise Exception('You cannot run pw.map() in the current state.'
                            ' Create a new pywren.ibm_cf_executor() instance.')

        map_futures, _ = self.executor.map(map_function=map_function,
                                           iterdata=map_iterdata,
                                           obj_chunk_size=chunk_size,
                                           extra_env=extra_env,
                                           extra_meta=extra_meta,
                                           remote_invocation=remote_invocation,
                                           remote_invocation_groups=remote_invocation_groups,
                                           invoke_pool_threads=invoke_pool_threads,
                                           data_all_as_one=data_all_as_one,
                                           overwrite_invoke_args=overwrite_invoke_args,
                                           exclude_modules=exclude_modules,
                                           job_max_runtime=timeout)
        self.futures.extend(map_futures)
        self._state = ExecutorState.running

        if len(map_futures) == 1:
            return map_futures[0]
        return map_futures

    def map_reduce(self, map_function, map_iterdata, reduce_function,
                   extra_env=None, extra_meta=None, chunk_size=None,
                   remote_invocation=False, timeout=wrenconfig.RUNTIME_TIMEOUT,
                   reducer_one_per_object=False, reducer_wait_local=False,
                   invoke_pool_threads=128, data_all_as_one=True,
                   overwrite_invoke_args=None, exclude_modules=None):
        """
        Map the map_function over the data and apply the reduce_function across all futures.
        This method is executed all within CF.
        :param map_function: the function to map over the data
        :param map_iterdata:  the function to reduce over the futures
        :param reduce_function:  the function to reduce over the futures
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.
        :param chunk_size: the size of the data chunks. 'None' for processing the whole file in one map
        :param data_type: the type of the data. Now allowed: None (files with newline) and csv.
        :param reducer_one_per_object: Set one reducer per object after running the partitioner
        :param reducer_wait_local: Wait for results locally
        :param invoke_pool_threads: Number of threads to use to invoke.
        :param data_all_as_one: upload the data as a single object. Default True
        :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
        :param exclude_modules: Explicitly keep these modules from pickled dependencies.
        :return: A list with size `len(map_iterdata)` of futures for each job

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map_reduce(foo, map_data_list, bar)
        """

        if self._state == ExecutorState.finished or self._state == ExecutorState.error:
            raise Exception('You cannot run pw.map_reduce() in the current state.'
                            ' Create a new pywren.ibm_cf_executor() instance.')

        map_futures, parts_per_object = self.executor.map(map_function, map_iterdata,
                                                          extra_env=extra_env,
                                                          extra_meta=extra_meta,
                                                          obj_chunk_size=chunk_size,
                                                          remote_invocation=remote_invocation,
                                                          invoke_pool_threads=invoke_pool_threads,
                                                          data_all_as_one=data_all_as_one,
                                                          overwrite_invoke_args=overwrite_invoke_args,
                                                          exclude_modules=exclude_modules,
                                                          job_max_runtime=timeout)

        self._state = ExecutorState.running
        if reducer_wait_local:
            self.monitor(futures=map_futures)

        futures = self.executor.reduce(reduce_function, map_futures, parts_per_object,
                                       reducer_one_per_object, extra_env, extra_meta)
        self.futures.extend(futures)

        if len(futures) == 1:
            return futures[0]
        return futures

    def monitor(self, futures=None, throw_except=True, return_when=ALL_COMPLETED,
                download_results=False, THREADPOOL_SIZE=128, WAIT_DUR_SEC=1):
        """
        Wait for the Future instances `fs` to complete. Returns a 2-tuple of
        lists. The first list contains the futures that completed
        (finished or cancelled) before the wait completed. The second
        contains uncompleted futures.

        :param return_when: One of `ALL_COMPLETED`, `ANY_COMPLETED`, `ALWAYS`
        :param THREADPOOL_SIZE: Number of threads to use. Default 64
        :param WAIT_DUR_SEC: Time interval between each check.
        :return: `(fs_dones, fs_notdones)`
            where `fs_dones` is a list of futures that have completed
            and `fs_notdones` is a list of futures that have not completed.
        :rtype: 2-tuple of lists

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map(foo, data_list)
          >>> dones, not_dones = pw.wait()
          >>> # not_dones should be an empty list.
          >>> results = [f.result() for f in dones]
        """
        if not futures:
            futures = self.futures

        if not futures:
            raise Exception('No activations to track. You must run pw.call_async(),'
                            ' pw.map() or pw.map_reduce() before call pw.wait()')

        msg = 'Executor ID {} Waiting for functions to complete'.format(self.executor_id)
        logger.info(msg)
        if not self.log_level and self._state == ExecutorState.running:
            print(msg)

        rabbit_amqp_url = None
        if self.rabbitmq_monitor:
            rabbit_amqp_url = self.config['rabbitmq'].get('amqp_url')

        if rabbit_amqp_url:
            logger.info('Going to use RabbitMQ to monitor function activations')

        pbar = None
        if not self.is_cf_cluster and not self.log_level \
           and return_when == ALL_COMPLETED and self._state == ExecutorState.running \
           and not is_notebook():
            import tqdm
            print()
            pbar = tqdm.tqdm(bar_format='  {l_bar}{bar}| {n_fmt}/{total_fmt}  ',
                             total=len(futures), disable=False)

        fs_dones, fs_notdones = wait(futures, self.executor_id, self.internal_storage,
                                     download_results=download_results,
                                     throw_except=throw_except, return_when=return_when,
                                     rabbit_amqp_url=rabbit_amqp_url, pbar=pbar,
                                     THREADPOOL_SIZE=THREADPOOL_SIZE, WAIT_DUR_SEC=WAIT_DUR_SEC)
        if pbar:
            pbar.close()
            print()

        self._state = ExecutorState.ready

        return fs_dones, fs_notdones

    def get_result(self, futures=None, throw_except=True, timeout=wrenconfig.RUNTIME_TIMEOUT,
                   THREADPOOL_SIZE=64, WAIT_DUR_SEC=2):
        """
        For getting PyWren results
        :param futures: Futures list. Default None
        :param throw_except: Reraise exception if call raised. Default True.
        :param verbose: Shows some information prints. Default False
        :param timeout: Timeout for waiting for results.
        :param THREADPOOL_SIZE: Number of threads to use. Default 64
        :return: The result of the future/s

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map(foo, data)
          >>> result = pw.get_result()
        """
        if futures:
            # Ensure futures is a list
            if type(futures) != list:
                ftrs = [futures]
            else:
                ftrs = futures
        else:
            # In this case self.futures is always a list
            ftrs = self.futures
            # self.futures = []

        if not ftrs:
            raise Exception('You must run pw.call_async(), pw.map()'
                            ' or pw.map_reduce() before call pw.get_result()')

        msg = 'Executor ID {} Getting results ...'.format(self.executor_id)
        logger.info(msg)
        if not self.log_level:
            print(msg)

        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(timeout)

        pbar = None
        if not self.is_cf_cluster and self._state != ExecutorState.ready \
           and not self.log_level and not is_notebook():
            import tqdm
            print()
            pbar = tqdm.tqdm(bar_format='  {l_bar}{bar}| {n_fmt}/{total_fmt}  ',
                             total=len(ftrs), disable=False)

        try:
            wait(ftrs, self.executor_id, self.internal_storage, download_results=True,
                 throw_except=throw_except, pbar=pbar, THREADPOOL_SIZE=THREADPOOL_SIZE,
                 WAIT_DUR_SEC=WAIT_DUR_SEC)
            result = [f.result() for f in ftrs if f.done and not f.futures]

        except TimeoutError:
            if pbar:
                pbar.close()
                print()
            not_dones_activation_ids = set([f.activation_id for f in ftrs if not f.done])
            msg = ('Executor ID {} Raised timeout of {} seconds getting results '
                   '\nActivations not done: {}'.format(self.executor_id, timeout, not_dones_activation_ids))
            logger.info(msg)
            if not self.log_level:
                print(msg)
            self._state = ExecutorState.error
            result = None

        except KeyboardInterrupt:
            if pbar:
                pbar.close()
                print()
            not_dones_activation_ids = [f.activation_id for f in ftrs if not f.done]
            msg = 'Executor ID {} Cancelled  \nActivations not done: {}'.format(self.executor_id, not_dones_activation_ids)
            logger.info(msg)
            if not self.log_level:
                print(msg)
            if self.data_cleaner and not self.is_cf_cluster:
                self.clean()
            exit()

        finally:
            signal.alarm(0)
            if pbar:
                pbar.close()
                print()
            if self.data_cleaner and not self.is_cf_cluster:
                self.clean()
            self._state = ExecutorState.result

        msg = "Executor ID {} Finished getting results".format(self.executor_id)
        logger.info(msg)
        if not self.log_level:
            print(msg)

        if result and len(result) == 1:
            return result[0]
        return result

    def create_timeline_plots(self, dst, name, run_statuses=None, invoke_statuses=None):
        """
        Creates timeline and histogram of the current execution in dst.

        :param dst: destination folder to save .png plots.
        :param name: name of the file.
        :param run_statuses: run statuses timestamps.
        :param invoke_statuses: invocation statuses timestamps.
        """
        from pywren_ibm_cloud.plots import create_timeline, create_histogram

        msg = 'Executor ID {} Creating timeline plots'.format(self.executor_id)
        logger.info(msg)
        if not self.log_level:
            print(msg)

        rabbitmq_used = self.rabbitmq_monitor

        if not run_statuses:
            if self._state == ExecutorState.new or self._state == ExecutorState.error:
                raise Exception('You must run pw.call_async(), pw.map() or pw.map_reduce()'
                                ' before call pw.create_timeline_plots()')

            if self._state == ExecutorState.running:
                # monitor() method not executed at any time
                self.monitor()
            if self._state == ExecutorState.ready:
                # wait() method already executed. Download statuses from storage
                self.rabbitmq_monitor = False
                self.monitor()

            if self.futures:
                run_statuses = [f.run_status for f in self.futures]
                invoke_statuses = [f.invoke_status for f in self.futures]
            else:
                logger.debug('No futures available to print the plots')
                return

            if rabbitmq_used and self.config['rabbitmq']['amqp_url'] and invoke_statuses:
                # delete download ststus timestamp
                for in_stat in invoke_statuses:
                    del in_stat['status_done_timestamp']

        create_timeline(dst, name, self.start_time, run_statuses, invoke_statuses)
        create_histogram(dst, name, self.start_time, run_statuses)

    def clean(self, local_execution=True):
        """
        Deletes all the files from COS. These files include the function,
        the data serialization and the function invocation results.
        """
        storage_bucket = self.config['pywren']['storage_bucket']
        storage_prerix = self.config['pywren']['storage_prefix']
        storage_prerix = os.path.join(storage_prerix, self.executor_id)

        msg = ("Executor ID {} Cleaning partial results from 'cos://{}/{}'".format(self.executor_id,
                                                                                   storage_bucket,
                                                                                   storage_prerix))
        logger.info(msg)
        if not self.log_level:
            print(msg)
            if not self.data_cleaner:
                print()

        if local_execution:
            # 1st case: Not background. The main code waits until the cleaner finishes its execution.
            # It is not ideal for performance tests, since it can take long time to complete.
            #clean_os_bucket(storage_bucket, storage_prerix, self.internal_storage)

            # 2nd case: Execute in Background as a subprocess. The main program does not wait for its completion.
            storage_config = json.dumps(self.internal_storage.get_storage_config())
            storage_config = storage_config.replace('"', '\\"')

            cmdstr = ("{} -c 'from pywren_ibm_cloud.storage.cleaner import clean_bucket; \
                              clean_bucket(\"{}\", \"{}\", \"{}\")'".format(sys.executable,
                                                                            storage_bucket,
                                                                            storage_prerix,
                                                                            storage_config))
            os.popen(cmdstr)

        else:
            extra_env = {'STORE_STATUS': False,
                         'STORE_RESULT': False}
            sys.stdout = open(os.devnull, 'w')
            self.executor.call_async(clean_os_bucket, [storage_bucket, storage_prerix], extra_env=extra_env)
            sys.stdout = sys.__stdout__

        self._state = ExecutorState.finished



import sys
import pkgutil


def main(args):
    runtime_meta = dict()
    mods = list(pkgutil.iter_modules())
    runtime_meta['preinstalls'] = [entry for entry in sorted([[mod, is_pkg] for _, mod, is_pkg in mods])]
    python_version = sys.version_info
    runtime_meta['python_ver'] = str(python_version[0])+"."+str(python_version[1])

    return runtime_meta



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os
import logging
from pywren_ibm_cloud.utils import version_str
from pywren_ibm_cloud.runtime import clone_runtime

logger = logging.getLogger(__name__)


def get_runtime_preinstalls(internal_storage, runtime, memory, config):
    """
    Download runtime information from storage at deserialize
    """
    log_level = os.getenv('PYWREN_LOG_LEVEL')
    logger.debug("Downloading runtime pre-installed modules from COS")
    try:
        runtime_meta = internal_storage.get_runtime_info('{}_{}'.format(runtime, memory))
        preinstalls = runtime_meta['preinstalls']
        if not log_level:
            print()
    except Exception:
        log_msg = 'Runtime {}_{} is not yet installed'.format(runtime, memory)
        logger.debug(log_msg)
        if not log_level:
            print('(Installing...)')
            old_stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')
        clone_runtime(runtime, memory=memory, config=config)
        if not log_level:
            sys.stdout = old_stdout
        runtime_meta = internal_storage.get_runtime_info('{}_{}'.format(runtime, memory))
        preinstalls = runtime_meta['preinstalls']

    if not runtime_valid(runtime_meta):
        raise Exception(("The indicated runtime: {} "
                         "is not appropriate for this Python version.")
                        .format(runtime))

    return preinstalls


def runtime_valid(runtime_meta):
    """
    Basic checks
    """
    logger.debug("Verifying Python versions")
    this_version_str = version_str(sys.version_info)
    return this_version_str == runtime_meta['python_ver']
